from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING

import gamspy._symbols.implicits as implicits
import gamspy.math
import gamspy.utils as utils
from gamspy._symbols.parameter import Parameter
from gamspy._symbols.set import Set
from gamspy._symbols.variable import Variable
from gamspy.exceptions import GamspyException, ValidationError

if TYPE_CHECKING:
    from gamspy import Container
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation
    from gamspy._symbols.alias import Alias
    from gamspy.math.misc import MathOp


def vector_norm(
    x: (
        Parameter
        | Variable
        | implicits.ImplicitParameter
        | implicits.ImplicitVariable
        | Expression
        | Operation
    ),
    ord: float | int = 2,
    dim: list[int] | list[Set | Alias] | None = None,
) -> Operation | Expression | MathOp:
    """
    Returns the vector norm of the provided vector x. If ord is not an even integer, absolute value is used which
    requires DNLP.

    Parameters
    ----------
    x : Parameter | Variable | implicits.ImplicitParameter | implicits.ImplicitVariable | Expression | Operation
    ord: int | float
    dim: list[int] | list[Set | Alias], optional

    Returns
    -------
    Expression | Operation | MathOp

    Examples
    --------
    >>> import gamspy as gp
    >>> import math
    >>> m = gp.Container()
    >>> i = gp.Set(m, name="i", records=["i1", "i2"])
    >>> vec = gp.Parameter(m, "vec", domain=[i], records=[("i1", 3), ("i2", 4)])
    >>> vlen = gp.Parameter(m, "vlen", domain=[])
    >>> vlen[...] = gp.math.vector_norm(vec)
    >>> math.isclose(vlen.toValue(), 5, rel_tol=1e-4)
    True

    """
    import gamspy._algebra.operation as operation
    from gamspy._symbols.alias import Alias

    if isinstance(ord, float):
        if ord.is_integer():
            ord = int(ord)
        elif ord in (float("inf"), float("-inf")):
            raise ValidationError("Infinity norms are not supported")

    if ord == 0:
        raise ValidationError("0 norm is not supported")

    even = isinstance(ord, int) and ord % 2 == 0
    domain = x.domain

    if len(domain) == 0:
        raise ValidationError("Provided argument is a scalar")

    sum_domain = domain
    if dim is not None:
        if not isinstance(dim, Iterable):
            raise ValidationError("dim must be an Iterable")

        if len(dim) == 0:
            raise ValidationError("If dim is provided, it must contain items")

        if isinstance(dim[0], int):
            for item in dim:
                if not isinstance(item, int):
                    raise ValidationError(
                        "If dim is provided, either all items must be integers"
                        " or all items must be Sets"
                    )

            sum_domain = []
            for d in dim:
                sum_domain.append(domain[d])
        else:
            for item in dim:
                if not isinstance(item, (Set, Alias)):
                    raise ValidationError(
                        "If dim is provided, either all items must be integers"
                        " or all items must be Sets"
                    )

            sum_domain = dim

    if ord == 2:
        return gamspy.math.sqrt(
            operation.Sum(sum_domain, gamspy.math.sqr(x[domain])),
            safe_cancel=True,
        )
    elif even:
        return gamspy.math.rpower(
            operation.Sum(sum_domain, x[domain] ** ord), (1 / ord)
        )
    elif ord == 1:
        return operation.Sum(sum_domain, gamspy.math.abs(x[domain]))

    return gamspy.math.rpower(
        operation.Sum(sum_domain, gamspy.math.abs(x[domain]) ** ord),
        (1 / ord),
    )


def next_alias(symbol: Alias | Set) -> Alias:
    """Provided the set or alias, it returns the next alias.
    If it is not found, it creates the alias. This function is
    mainly for matrix multiplication conflict resolution but
    it might be helpful in the cases where you need to generate
    many aliases from a set.

    Parameters
    ----------
    symbol : Set | Alias

    Returns
    -------
    Alias

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, name="i", records=["i1", "i2", "i3"])
    >>> j = gp.math.next_alias(i)
    >>> j.name
    'AliasOfi_2'
    >>> k = gp.math._generate_dims(m, [10])[0]
    >>> k.name
    'DenseDim10_1'
    >>> k2 = gp.math.next_alias(k)
    >>> k2.name
    'DenseDim10_2'

    """

    current = symbol
    if symbol.name.startswith("DenseDim") or symbol.name.startswith("AliasOf"):
        prefix, num = symbol.name.split("_")
    else:
        prefix, num = f"AliasOf{symbol.name}", 1

    num = int(num) + 1
    expected_name = f"{prefix}_{num}"
    find_x = symbol.container.data.get(expected_name, None)
    if find_x is None:
        find_x = symbol.container.addAlias(expected_name, alias_with=current)

    return find_x


@dataclass
class Dim:
    dims: list[int]


def dim(dims: list[int] | tuple[int, ...]) -> Dim:
    """Returns an array where each element
    corresponds to a set where the dimension of the
    set is equal to the element in dims. If same dimension
    size used, then next free alias will be returned.
    Symbols are generated once the Dim object is passed to
    a constructor that supports it.

    Parameters
    ----------
    dims: list[int] | tuple[int, ...]

    Returns
    -------
    Dim

    Examples
    --------
    >>> import gamspy as gp
    >>> import math
    >>> m = gp.Container()
    >>> a = gp.math.dim([10, 20]) # nothing generated yet
    >>> a
    Dim(dims=[10, 20])
    >>> par = gp.Parameter(m, name="par", domain=a) # now two sets are generated
    >>> par.domain
    [Set(name='DenseDim10_1', domain=['*']), Set(name='DenseDim20_1', domain=['*'])]
    >>> par2 = gp.Parameter(m, name="par2", domain=a) # same 2 sets are used
    >>> par2.domain
    [Set(name='DenseDim10_1', domain=['*']), Set(name='DenseDim20_1', domain=['*'])]

    """
    for x in dims:
        if not isinstance(x, int):
            raise ValidationError("Dimensions must be integers")

    return Dim(dims=dims)  # type: ignore


def _generate_dims(
    m: Container, dims: list[int] | tuple[int, ...]
) -> list[Alias | Set]:
    sets_so_far = []
    for x in dims:
        expected_name = f"DenseDim{x}_1"
        find_x = m.data.get(expected_name, None)
        if find_x is None:
            find_x = m.addSet(
                name=expected_name, records=range(x), is_singleton=(x == 1)
            )

        while find_x in sets_so_far:
            find_x = next_alias(find_x)

        sets_so_far.append(find_x)

    return sets_so_far


def trace(
    x: (
        Parameter | implicits.ImplicitParameter | Variable | implicits.ImplicitVariable
    ),
    axis1: int = 0,
    axis2: int = 1,
) -> Operation:
    """Returns trace of the given input x.
    By default trace of zeroth and first axis used. `axis1` and `axis2` parameters
    control on which axes to get trace. Domains at the axis1 and axis2 must be same
    or aliases.


    Parameters
    ----------
    x: (
        Parameter
        | implicits.ImplicitParameter
        | Variable
        | implicits.ImplicitVariable
    )
    axis1=0
    axis2=1

    Returns
    -------
    Operation

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> m = gp.Container()
    >>> identity = np.eye(3, 3)
    >>> mat = gp.Parameter(m, name="mat", domain=gp.math.dim([3, 3]), records=identity, uels_on_axes=True)
    >>> sc = gp.Parameter(m, name="sc", domain=[])
    >>> sc[...] = gp.math.trace(mat)
    >>> int(sc.toDense())
    3

    """
    import gamspy._algebra.operation as operation

    if len(x.domain) < 2:
        raise ValidationError("Trace requires at least 2 dimensions")

    if not utils.setBaseEqual(x.domain[axis1], x.domain[axis2]):
        raise ValidationError("Matrix dimensions are not equal")

    domain = list(x.domain)
    domain[axis1] = domain[axis2]

    return operation.Sum(domain[axis2], x[domain])


def permute(
    x: (
        Parameter | implicits.ImplicitParameter | Variable | implicits.ImplicitVariable
    ),
    dims: list[int],
) -> implicits.ImplicitVariable | implicits.ImplicitParameter:
    """Permutes the dimensions provided input `x` using `dim`.
    Similar to PyTorch permute.

    Parameters
    ----------
    x: (
        Parameter
        | implicits.ImplicitParameter
        | Variable
        | implicits.ImplicitVariable
    )
    dims: list[int]

    Returns
    -------
    implicits.ImplicitVariable | implicits.ImplicitParameter

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, name="i")
    >>> j = gp.Set(m, name="j")
    >>> k = gp.Set(m, name="k")
    >>> p = gp.Parameter(m, name="p", domain=[i, j, k])
    >>> p2 = gp.math.permute(p, [2, 0, 1])
    >>> p2.domain
    [Set(name='k', domain=['*']), Set(name='i', domain=['*']), Set(name='j', domain=['*'])]

    """
    # TODO Accept permuting expressions!
    # Might be needed in some context
    for i in dims:
        if not isinstance(i, int):
            raise ValidationError("Permute dimensions must be integers")

    dims_len = len(dims)
    if min(dims) != 0 or max(dims) != dims_len - 1:
        raise ValidationError("Permute requires the order of indices from 0 to n-1")

    if len(set(dims)) != dims_len:
        raise ValidationError("Permute dimensions must be unique")

    permuted_domain = utils._permute_domain(x.domain, dims)
    if isinstance(x, Parameter):
        return implicits.ImplicitParameter(
            x,
            name=x.name,
            records=x.records,
            domain=permuted_domain,
            permutation=dims,
        )
    elif isinstance(x, implicits.ImplicitParameter):
        if x.permutation is not None:
            dims = utils._permute_domain(x.permutation, dims)

        return implicits.ImplicitParameter(
            x.parent,
            name=x.name,
            records=x._records,
            domain=permuted_domain,
            permutation=dims,
            scalar_domains=x._scalar_domains,
        )
    elif isinstance(x, Variable):
        return implicits.ImplicitVariable(
            x, name=x.name, domain=permuted_domain, permutation=dims
        )
    elif isinstance(x, implicits.ImplicitVariable):
        if x.permutation is not None:
            dims = utils._permute_domain(x.permutation, dims)

        return implicits.ImplicitVariable(
            x.parent,
            name=x.name,
            domain=permuted_domain,
            permutation=dims,
            scalar_domains=x._scalar_domains,
        )

    raise GamspyException(f"permute not implemented for {type(x)}")


def _validate_matrix_mult_dims(left, right):
    """Validates the dimensions for the matrix multiplication"""
    left_len = len(left.domain)
    right_len = len(right.domain)

    dim_no_match_err = "Matrix multiplication dimensions do not match"

    if left_len == 0:
        raise ValidationError(
            "Matrix multiplication requires at least 1 domain, left side is a scalar"
        )

    if right_len == 0:
        raise ValidationError(
            "Matrix multiplication requires at least 1 domain, right side is a scalar"
        )

    lr = (left_len, right_len)

    left_controlled = getattr(left, "controlled_domain", [])
    right_controlled = getattr(right, "controlled_domain", [])
    controlled_domain = [*left_controlled, *right_controlled]

    if lr == (1, 1):
        # Dot product
        if not utils.setBaseEqual(left.domain[0], right.domain[0]):
            raise ValidationError("Dot product requires same domain")

        sum_domain = left.domain[0]
        while sum_domain in controlled_domain:
            sum_domain = next_alias(sum_domain)

        return [sum_domain], [sum_domain], sum_domain
    elif lr == (2, 2):
        # Matrix multiplication
        if not utils.setBaseEqual(left.domain[1], right.domain[0]):
            raise ValidationError(dim_no_match_err)

        left_domain = left.domain[0]
        right_domain = right.domain[1]
        if left_domain == right_domain:
            left_domain = next_alias(left_domain)

        sum_domain = left.domain[1]
        while (
            sum_domain in (left_domain, right_domain) or sum_domain in controlled_domain
        ):
            sum_domain = next_alias(sum_domain)

        return (
            [left_domain, sum_domain],
            [sum_domain, right_domain],
            sum_domain,
        )
    elif lr == (1, 2):
        # Vector matrix, vector 1-prepended
        if not utils.setBaseEqual(left.domain[0], right.domain[0]):
            raise ValidationError(dim_no_match_err)

        sum_domain = left.domain[0]
        right_domain = right.domain[1]

        while right_domain in controlled_domain:
            right_domain = next_alias(right_domain)

        while sum_domain == right_domain or sum_domain in controlled_domain:
            sum_domain = next_alias(sum_domain)

        return [sum_domain], [sum_domain, right_domain], sum_domain
    elif lr == (2, 1):
        # Matrix vector, ordinary
        if not utils.setBaseEqual(left.domain[1], right.domain[0]):
            raise ValidationError(dim_no_match_err)

        sum_domain = left.domain[1]
        while left.domain[0] == sum_domain or sum_domain in controlled_domain:
            sum_domain = next_alias(sum_domain)

        return [left.domain[0], sum_domain], [sum_domain], sum_domain
    elif left_len == 1 and right_len > 2:
        # Vector batched-matrix, vector 1-prepended
        if not utils.setBaseEqual(left.domain[0], right.domain[-2]):
            raise ValidationError(dim_no_match_err)

        sum_domain = left.domain[0]
        while (
            sum_domain in right.domain[:-2]
            or sum_domain == right.domain[-1]
            or sum_domain in controlled_domain
        ):
            sum_domain = next_alias(sum_domain)

        return (
            [sum_domain],
            [*right.domain[:-2], sum_domain, right.domain[-1]],
            sum_domain,
        )
    elif left_len > 2 and right_len == 1:
        # batched-matrix vector, ordinary
        if not utils.setBaseEqual(left.domain[-1], right.domain[0]):
            raise ValidationError(dim_no_match_err)

        sum_domain = left.domain[-1]
        while sum_domain in left.domain[:-1] or sum_domain in controlled_domain:
            sum_domain = next_alias(sum_domain)

        return (
            [*left.domain[:-1], sum_domain],
            [sum_domain],
            sum_domain,
        )
    elif left_len >= 2 and right_len >= 2:
        # batched-matrix batched-matrix
        if not utils.setBaseEqual(left.domain[-1], right.domain[-2]):
            raise ValidationError(dim_no_match_err)

        batch_dim_1 = left.domain[:-2]
        batch_dim_2 = right.domain[:-2]

        if len(batch_dim_1) > 0 and len(batch_dim_2) > 0:
            if len(batch_dim_1) != len(batch_dim_2):
                raise ValidationError("Batch dimensions do not match")

            if any(x != y for x, y in zip(batch_dim_1, batch_dim_2, strict=False)):
                raise ValidationError("Batch dimensions do not match")

        left_domain = left.domain[-2]
        right_domain = right.domain[-1]
        while right_domain == left_domain or right_domain in batch_dim_1:
            right_domain = next_alias(right_domain)

        while left_domain == right_domain or left_domain in batch_dim_2:
            left_domain = next_alias(left_domain)

        sum_domain = left.domain[-1]
        while (
            sum_domain in left.domain[:-1]
            or sum_domain in right.domain[:-2]
            or sum_domain in (right_domain, left_domain)
            or sum_domain in controlled_domain
        ):
            sum_domain = next_alias(sum_domain)

        return (
            [*left.domain[:-2], left_domain, sum_domain],
            [*right.domain[:-2], sum_domain, right_domain],
            sum_domain,
        )

    raise ValidationError(
        f"Matrix multiplication for left dim: {left_len},"
        f" right dim: {right_len} not implemented"
    )



"""Exception classes for GAMSPy"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gamspy import Options


class FatalError(Exception):
    """Indicates an error that cannot be recovered from. This error should never be caught."""


class GamspyException(Exception):
    """
    Plain Gamspy exception. This exception can be caught and GAMSPy should be able to continue.

    Parameters
    ----------
    message : str
    return_code : int | None
    """

    def __init__(self, message: str, return_code: int | None = None) -> None:
        super().__init__(message)
        self.message = message
        self.return_code = return_code


class ValidationError(Exception):
    """An error while validating data."""


class NeosClientException(Exception):
    """NeosClient exception."""


class EngineClientException(Exception):
    """EngineClient exception"""


class NeosException(GamspyException):
    """NEOS Server execution exception"""


class LatexException(GamspyException):
    """To latex execution exception"""


class EngineException(GamspyException):
    """
    GAMS Engine execution exception.

    Parameters
    ----------
    message : str
    return_code : int
    status_code : int | None, optional
    """

    def __init__(
        self,
        message: str,
        return_code: int,
        status_code: int | None = None,
    ) -> None:
        super().__init__(message, return_code)
        self.status_code = status_code


CHUNK_SIZE = 8
FRAME = "=" * 13


def _parse_errors(lines: list[str], index: int) -> str:
    error_lines = [lines[index - 1]]
    temp_index = index

    try:
        chunk = lines[temp_index : temp_index + CHUNK_SIZE]

        while any("****" in err_line for err_line in chunk):
            error_lines.extend(chunk)
            temp_index += CHUNK_SIZE
            chunk = lines[temp_index : temp_index + CHUNK_SIZE]
    except IndexError:
        ...

    error_message = f"\n\n{FRAME}\nError Summary\n{FRAME}\n{''.join(error_lines)}"

    return error_message.rstrip()


def _customize_exception(
    options: Options,
    job_name: str,
    return_code: int | None,
) -> str:
    error_message = ""
    if options.write_listing_file is False or return_code is None:
        return ""

    if options.listing_file:
        lst_path = (
            options.listing_file
            if os.path.isabs(options.listing_file)
            else os.path.join(os.getcwd(), options.listing_file)
        )
    else:
        lst_path = job_name + ".lst"

    with open(lst_path, encoding="utf-8") as lst_file:
        all_lines = lst_file.readlines()

    num_lines = len(all_lines)
    index = 0
    while index < num_lines:
        line = all_lines[index]

        if line.startswith("****"):
            error_message = _parse_errors(all_lines, index)
            break

        index += 1

    return error_message



from __future__ import annotations

import math

import numpy as np

import gamspy as gp
from gamspy.exceptions import ValidationError


def _encode_infinity(x: np.ndarray) -> np.ndarray:
    """
    Encode infinity values as complex numbers.
    - Replace -np.inf with 0 - 1j.
    - Replace np.inf with 0 + 1j.
    """
    x = np.where(x == -np.inf, 0 - 1j, x)
    x = np.where(x == np.inf, 0 + 1j, x)
    return x


def _decode_complex_array(z: np.ndarray) -> np.ndarray:
    """
    Decode complex numbers back to infinities
    - Replace 0 - 1j -> -np.inf
    - Replace 0 + 1j -> np.inf
    """
    real = z.real.copy()
    real[z.imag > 0] = np.inf
    real[z.imag < 0] = -np.inf
    return real


def _generate_name(sym_type: str, prefix: str, name: str) -> str:
    rand = gp.utils._get_unique_name()[:5]
    return "_".join([sym_type, prefix, name, rand])


def _check_tuple_int(
    value: int | tuple[int] | tuple[int, int],
    name: str,
    allow_zero=False,
) -> tuple[int, int]:
    if not isinstance(value, (int, tuple)):
        raise ValidationError(f"{name} must be an integer or a tuple of integer")

    if isinstance(value, int):
        value = (value, value)
    elif len(value) == 1:
        value = (value[0], value[0])

    cmp = (lambda a: a >= 0) if allow_zero else (lambda a: a > 0)
    text = "or equal to " if allow_zero else ""

    if not (isinstance(value[0], int) and cmp(value[0])):
        raise ValidationError(f"{name} must be greater than {text}0")

    if not (isinstance(value[1], int) and cmp(value[1])):
        raise ValidationError(f"{name} must be a greater than {text}0")

    return value


def _check_padding(value: int | tuple[int, int]) -> tuple[int, int, int, int]:
    if not isinstance(value, (int, tuple)):
        raise ValidationError("Padding must be an integer or a tuple of two integers")

    if isinstance(value, int):
        padding = (value, value, value, value)

    if isinstance(value, tuple):
        if len(value) != 2:
            raise ValidationError(
                f"Padding can only be a tuple of 2 integers. Not {len(value)}."
            )

        # padding is represented as (top, left, bottom, right)
        padding = (value[0], value[1], value[0], value[1])

    if not (isinstance(padding[0], int) and (padding[0] >= 0)):
        raise ValidationError("Padding must be greater than or equal to 0")

    if not (isinstance(padding[1], int) and (padding[1] >= 0)):
        raise ValidationError("Padding must be a greater than or equal to 0")

    return padding


def _calc_same_padding_2d(
    kernel_size: tuple[int, int],
) -> tuple[int, int, int, int]:
    # assumes stride = 1
    pad_h_total = max(kernel_size[0] - 1, 0)
    pad_w_total = max(kernel_size[1] - 1, 0)

    # Calculate padding for height
    pad_top = pad_h_total // 2
    pad_bottom = pad_h_total - pad_top

    # Calculate padding for width
    pad_left = pad_w_total // 2
    pad_right = pad_w_total - pad_left

    return (pad_top, pad_left, pad_bottom, pad_right)


def _calc_same_padding_1d(
    kernel_size: int,
) -> tuple[int, int]:
    # assumes stride = 1
    pad_w_total = max(kernel_size - 1, 0)

    # Calculate padding for width
    pad_left = pad_w_total // 2
    pad_right = pad_w_total - pad_left

    return (pad_left, pad_right)


def _calc_w(
    padding: tuple[int, int] | str,
    kernel_size: int,
    stride: int,
    w_in: int,
) -> int:
    # same padding
    if isinstance(padding, str):
        return w_in

    w_out = math.floor(
        1 + ((w_in + (padding[0] + padding[1]) - (kernel_size - 1) - 1) / stride)
    )

    return w_out


def _calc_hw(
    padding: tuple[int, int, int, int] | tuple[int, int] | str,
    kernel_size: tuple[int, int],
    stride: tuple[int, int],
    h_in: int,
    w_in: int,
) -> tuple[int, int]:
    if isinstance(padding, str):
        return h_in, w_in

    if len(padding) == 2:
        # padding is represented as (top, left, bottom, right)
        padding = (padding[0], padding[1], padding[0], padding[1])

    h_out = math.floor(
        1 + ((h_in + (padding[0] + padding[2]) - (kernel_size[0] - 1) - 1) / stride[0])
    )
    w_out = math.floor(
        1 + ((w_in + (padding[1] + padding[3]) - (kernel_size[1] - 1) - 1) / stride[1])
    )

    return h_out, w_out


def _next_domains(
    input_domain: list[gp.Alias | gp.Set], check_domains: list[gp.Set]
) -> list[gp.Set]:
    names = {x.name for x in check_domains}
    output = []
    for domain in input_domain:
        while domain.name in names:
            domain = gp.math.next_alias(domain)

        output.append(domain)
        names.add(domain.name)

    return output



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._algebra.condition as condition
import gamspy.utils as utils
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Set
    from gamspy._symbols.implicits import ImplicitSet


class Domain:
    """
    Domain class needed for where statements on multidimensional index list
    in operations

    Parameters
    ----------
    sets: tuple[Set | str]

    Examples
    --------
    >>> from gamspy import Container, Set, Ord, Card, Variable, Equation, Sum, Domain
    >>> m = Container()
    >>> X = Set(m, name="X", records=[f"I{i}" for i in range(1, 22)])
    >>> Y = Set(m, name="Y", records=[f"J{j}" for j in range(1, 22)])
    >>> inside = Set(m, name="inside", domain=[X, Y])
    >>> inside[X, Y].where[~((Ord(X) == 1) & (Ord(X) == Card(X)))] = True
    >>> f = Variable(m, name="f", domain=[X, Y], type="positive")
    >>> obj = Variable(m, name="obj")
    >>> objfun = Equation(m, name="objfun", type="regular")
    >>> objfun[...] = obj == Sum(Domain(X, Y).where[inside[X, Y]], f[X.lead(1), Y] - f[X, Y])

    """

    def __init__(self, *sets: Set | Alias | ImplicitSet) -> None:
        self._sanity_check(sets)
        self.sets = sets
        self.container = self._find_container()  # type: ignore
        self.domain = sets
        self.where = condition.Condition(self)

    def __repr__(self) -> str:
        return f"Domain(sets={self.sets})"

    def _sanity_check(self, sets: tuple[Set | Alias | ImplicitSet, ...]):
        if len(sets) < 2:
            error_message = f"Domain requires at least 2 sets but found {len(sets)}. "
            if len(sets) == 1:
                error_message += f"You can directly limit the domain with {sets[0].name}.where[<your_limitation>]."
            raise ValidationError(error_message)

        if all(not hasattr(set, "container") for set in sets):
            raise ValidationError(
                "At least one of the sets in the domain must be a Set or Alias"
            )

    def _find_container(self):
        for set in self.sets:
            if hasattr(set, "container"):
                return set.container

    def gamsRepr(self) -> str:
        """
        Representation of this Domain in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Domain
        >>> m = Container()
        >>> X = Set(m, name="x", records=[f"I{i}" for i in range(1, 22)])
        >>> Y = Set(m, name="y", records=[f"J{j}" for j in range(1, 22)])
        >>> Domain(X, Y).gamsRepr()
        '(x,y)'

        """
        return utils._get_domain_str(self.sets)

    def latexRepr(self) -> str:
        """
        Representation of this Domain in Latex.

        Returns
        -------
        str
        """
        return utils._get_domain_str(self.sets, latex=True)[1:-1]



from gamspy.formulations.ml.decision_tree_struct import DecisionTreeStruct
from gamspy.formulations.ml.gradient_boosting import GradientBoosting
from gamspy.formulations.ml.random_forest import RandomForest
from gamspy.formulations.ml.regression_tree import RegressionTree

__all__ = [
    "DecisionTreeStruct",
    "GradientBoosting",
    "RandomForest",
    "RegressionTree",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from gamspy.math.misc import MathOp

if TYPE_CHECKING:
    from gamspy._algebra.expression import Expression
    from gamspy._types import OperableType


def binomial(n: OperableType, k: OperableType) -> MathOp:
    """
    (Generalized) Binomial coefficient for ``n > -1`` and ``-1 < k < n + 1``

    Parameters
    ----------
    n : OperableType
    k : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import binomial
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> p = Parameter(m, "p", domain=i, records=[("i1", 0.3), ("i2", 0.8), ("i3", 0.45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = binomial(75, p[i])

    """
    return MathOp("binomial", (n, k))


def centropy(
    x: OperableType,
    y: OperableType,
    z: float = 1e-20,
) -> MathOp:
    """
    Cross-entropy: ``x.ln((x + z) / (y + z))`` for ``x, y > 0`` and ``z >= 0``

    Parameters
    ----------
    x : OperableType
    y : OperableType
    z : float, optional

    Returns
    -------
    MathOp

    Raises
    ------
    ValueError
        if z is smaller than 0

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import centropy
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> x = Parameter(m, "x", domain=i, records=[("i1", 0.3), ("i2", 8), ("i3", 45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = centropy(2.8, x[i])
    >>> b.toList()
    [('i1', 6.254058220219863), ('i2', -2.939501948596297), ('i3', -7.775720603249651)]

    """
    if not isinstance(z, (int, float)):
        raise TypeError("z must be a number")

    if z < 0:
        raise ValueError("z must be greater than or equal to 0")

    return MathOp("centropy", (x, y, z))


def uniform(
    lower_bound: float | Expression,
    upper_bound: float | Expression,
) -> MathOp:
    """
    Generates a random number from the uniform distribution between
    ``lower_bound`` and ``higher_bound``

    Parameters
    ----------
    lower_bound : float
    upper_bound : float

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import uniform
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> x = Parameter(m, "x", domain=i, records=[("i1", 30), ("i2", 8), ("i3", 45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = uniform(x[i], 50)

    """
    return MathOp("uniform", (lower_bound, upper_bound))


def uniformInt(lower_bound: int | float, upper_bound: int | float) -> MathOp:
    """
    Generates an integer random number from the discrete uniform distribution
    whose outcomes are the integers between ``lower_bound`` and ``higher_bound``

    Parameters
    ----------
    lower_bound : int | float
    upper_bound : int | float

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import uniformInt
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> x = Parameter(m, "x", domain=i, records=[("i1", 30), ("i2", 8), ("i3", 45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = uniformInt(x[i], 50)
    >>> b.toList()
    [('i1', 33.0), ('i2', 44.0), ('i3', 48.0)]

    """
    return MathOp("uniformInt", (lower_bound, upper_bound))


def normal(mean: int | float, dev: int | float) -> MathOp:
    """
    Generate a random number from the normal distribution with mean ``mean``
    and standard deviation ``dev``

    Parameters
    ----------
    mean : int | float
    dev : int | float

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import normal
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> x = Parameter(m, "x", domain=i, records=[("i1", 30), ("i2", 8), ("i3", 45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = normal(x[i], 5)

    """
    return MathOp("normal", (mean, dev))



from __future__ import annotations

import os
import threading
from typing import TYPE_CHECKING

import gams.transfer as gt
from gams.core.gdx import GMS_DT_ALIAS

import gamspy as gp
import gamspy._algebra.condition as condition
import gamspy._validation as validation
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Container


class UniverseAlias(gt.UniverseAlias):
    """
    Represents a UniverseAlias symbol in GAMS.

    Parameters
    ----------
    container : Container
    name : str

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> universe = gp.UniverseAlias(m)
    >>> universe.records
    Empty DataFrame
    Columns: [uni]
    Index: []
    >>> i = gp.Set(m, "i", records=['i1', 'i2'])
    >>> universe.records
      uni
    0  i1
    1  i2

    """

    @classmethod
    def _constructor_bypass(cls, container: Container, name: str):
        # create new symbol object
        obj = UniverseAlias.__new__(cls, container, name)

        # set private properties directly
        obj._requires_state_check = False
        obj._container = container
        container._requires_state_check = True
        obj._name = name
        obj._modified = True

        # typing
        obj._gams_type = GMS_DT_ALIAS
        obj._gams_subtype = 0

        # add to container
        container.data.update({name: obj})

        # gamspy attributes
        obj.where = condition.Condition(obj)
        obj._latex_name = name.replace("_", r"\_")

        # add statement
        obj.container._add_statement(obj)

        return obj

    def __new__(cls, container: Container | None = None, name: str = "universe"):
        if container is not None and not isinstance(container, gp.Container):
            raise TypeError(
                f"Container must of type `Container` but found {type(container)}"
            )

        if not isinstance(name, str):
            raise TypeError(f"Name must of type `str` but found {type(name)}")

        try:
            if not container:
                container = gp._ctx_managers[(os.getpid(), threading.get_native_id())]

            symbol = container[name]
            if isinstance(symbol, cls):
                return symbol

            raise TypeError(
                f"Cannot overwrite symbol `{name}` in container"
                " because it is not a UniverseAlias object)"
            )
        except KeyError:
            return object.__new__(cls)

    def __init__(self, container: Container | None = None, name: str = "universe"):
        # check if the name is a reserved word
        name = validation.validate_name(name)
        if container is None:
            try:
                container = gp._ctx_managers[(os.getpid(), threading.get_native_id())]
            except KeyError as e:
                raise ValidationError("UniverseAlias requires a container.") from e

        super().__init__(container, name)
        self._latex_name = self.name.replace("_", r"\_")

        # allow conditions
        self.where = condition.Condition(self)

        self.container._add_statement(self)
        self.container._synch_with_gams()

    def _serialize(self) -> dict:
        return {}

    def _deserialize(self, info: dict) -> None: ...

    def __repr__(self) -> str:
        return f"UniverseAlias(name='{self.name}')"

    def gamsRepr(self) -> str:
        """
        Representation of the UniverseAlias in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.UniverseAlias(m, name="universe")
        >>> i.gamsRepr()
        'universe'

        """
        return self.name

    def latexRepr(self) -> str:
        """
        Representation of the UniverseAlias in LaTeX.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.UniverseAlias(m, name="universe_alias")
        >>> i.latexRepr()
        'universe\\_alias'

        """
        return self._latex_name

    def getDeclaration(self) -> str:
        """
        Declaration of the UniverseAlias in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.UniverseAlias(m, name="universe")
        >>> i.getDeclaration()
        'Alias(universe,*);'

        """
        return f"Alias({self.name},*);"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Sequence

    import gamspy as gp

    MatchesType: TypeAlias = dict[
        gp.Equation | Sequence[gp.Equation],
        gp.Variable | Sequence[gp.Variable],
    ]


class FormulationResult:
    """
    FormulationResult class provides a common interface for returning results
    when formulations are called. In the old convention, formulations returned
    a tuple of result variable and list of equations. In some cases it was
    possible to get extra output from the formulation. To provide backwards
    compatibility, FormulationResult class can be unpacked into a result
    variable and list of equations. Also it supports returning extra output in
    unpacking.

    With the FormulationResult you can have more access to underlying symbols
    created such as equations, variables, parameters and sets. Since many
    formulations created symbols with randomized names, it was tedious to find
    intermediate symbols created. FormulationResult has dictionaries where keys
    are expected to be documented in the formulation returning the
    FormulationResult therefore you can access a symbol via its known key.

    For example:

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> x = gp.Variable(m)
    >>> res = gp.math.activation.relu_with_binary_var(x)
    >>> aux_binary_var = res.variables_created["binary"]

    Therefore, it is important for the formulation returning a
    FormulationResult to properly list the keys to the symbols that are
    created.

    FormulationResult has the following attributes that might be useful:
        - `result`
        - `equations_created`
        - `sets_created`
        - `parameters_created`
        - `matches`
        - `other`
        - `extra_return`

    """

    def __init__(
        self,
        result: gp.Variable | gp.Parameter | None = None,
        equations_created: dict[str, gp.Equation] | None = None,
        extra_return: gp.Variable | MatchesType | None = None,
    ) -> None:
        self.result = result
        self.equations_created: dict[str, gp.Equation] = (
            equations_created if equations_created is not None else {}
        )

        self.variables_created: dict[str, gp.Variable] = {}
        self.sets_created: dict[str, gp.Set] = {}
        self.parameters_created: dict[str, gp.Parameter] = {}
        self.matches: MatchesType = {}
        self.other: dict[str, Any] = {}

        # Extra return is only here for backwards compat reasons
        # try to use other when you can
        self.extra_return = extra_return

    def __len__(self):
        # For backwards compat
        return 2 + (self.extra_return is not None)

    def __iter__(self):
        # For backwards compat
        if self.extra_return is None:
            return [
                self.result,
                list(self.equations_created.values()),
            ].__iter__()

        return [
            self.result,
            self.extra_return,
            list(self.equations_created.values()),
        ].__iter__()

    def __str__(self) -> str:
        return (
            "FormulationResult(\n"
            + "    Equations: "
            + str(self.equations_created.keys())
            + "\n"
            + "    Variables: "
            + str(self.variables_created.keys())
            + "\n"
            + "    Sets: "
            + str(self.sets_created.keys())
            + "\n"
            + "    Parameters: "
            + str(self.parameters_created.keys())
            + "\n"
            + "    Num Matches: "
            + str(len(self.matches))
            + "\n"
            + "    Other: "
            + str(self.other)
            + "\n"
            + ")"
        )
